"use strict";
var util = require('./util');
var OBD = require('obd-parser');
var Promise = require('bluebird');
var uuid = require('uuid');
var zlib_1 = require('zlib');
var path_1 = require('path');
var fs_1 = require('fs');
var connection_1 = require('./connection');
var mkdirp = require('mkdirp');
var PQueue = require('p-queue');
// Each trip needs a unique identifier so we can avoid fuzzy matching times
var TRIP_UUID = uuid.v4();
// We want to keep file sizes in check
var MAX_FILE_SIZE = (128 * 1024);
// The pollers that are being requested by the monitor
var pollers = [];
// Stream used to write output data to disk/stdout
var outputStream;
// Zip stream...if we are using one
var zipper;
// Used to queue writes of OBD data to output file(s)
var pqueue = new PQueue({
    concurrency: 1
});
/**
 * Collates all PIDs and their desired intervals, then begins the monitor loop
 * @return {void}
 */
function default_1(pids) {
    Promise.map(pids, function (pidCodeOrName) {
        var pid = pidCodeOrName.split(':')[0];
        var interval = parseInt(pidCodeOrName.split(':')[1], 10);
        if (!pid || !interval || isNaN(interval)) {
            console.error("PID value " + pidCodeOrName + " is not valid. Format must be PID:INTERVAL, e.g 0C:500 to get RPM every 500 milliseconds");
            process.exit(1);
        }
        var monitor = {
            pid: util.getPidByCode(pid) || util.getPidByName(pid),
            interval: interval
        };
        return monitor;
    })
        .then(function (pids) { return monitor(pids); });
}
exports.__esModule = true;
exports["default"] = default_1;
/**
 * Creates the absolute path for an output file directory if the output
 * option is passed.
 * @return {String}
 */
function getOutputDirectoryName() {
    var outdir = path_1.join(util.getProgram().outdir, new Date().toJSON().split('T')[0], TRIP_UUID);
    return path_1.isAbsolute(outdir) ? outdir : path_1.join(process.cwd(), outdir);
}
/**
 * Creates a nested directory for storage of monitor results
 * @return {Promise<any>}
 */
function createOutputDir(p) {
    return new Promise(function (resolve, reject) {
        mkdirp(p, function (err) {
            if (err) {
                reject(err);
            }
            else {
                resolve();
            }
        });
    });
}
/**
 * Function for Array.prototype.sort for monitor files sorting.
 * Essentialy an integer sort.
 * @return {Number}
 */
function fileSorter(a, b) {
    return parseInt(a, 10) - parseInt(b, 10);
}
/**
 * Function for Array.prototype.sort for monitor files sorting.
 * Essentialy an integer sort.
 * @return {Number}
 */
function fileFilter(a) {
    return !isNaN(parseInt(a, 10));
}
/**
 * Get the extension to use for files created by our monitor
 * @return {String}
 */
function getFileExt() {
    return util.getProgram().zip ? '.json.zip' : '.json';
}
/**
 * Returns the next filename that will be used to store OBD output.
 * These are numeric and ascend from 0..n
 * @return {Promise<string>}
 */
function getNextFilename(p) {
    return new Promise(function (resolve, reject) {
        fs_1.readdir(p, function (err, list) {
            if (err) {
                reject(err);
            }
            else {
                if (list.length === 0) {
                    resolve("0" + getFileExt());
                }
                else {
                    var largestFilename = list.filter(fileFilter).sort(fileSorter).pop();
                    var n = parseInt(path_1.basename(largestFilename || '0', '.json'), 10);
                    resolve("" + (n + 1) + getFileExt());
                }
            }
        });
    });
}
/**
 * Returns the writeable stream that OBD data will be written to. This also
 * handles determining if file should be rotated and if we need to return a
 * zip stream vs. a plain fs stream.
 * @return {Promise<NodeJS.WritableStream>}
 */
function getWriteStream() {
    var outdir = util.getProgram().outdir;
    var useZip = util.getProgram().zip;
    // User wants to print to stdout, easy!
    if (!outdir) {
        return Promise.resolve(outputStream = process.stdout);
    }
    // Existing file stream has not exceeded file size so keep writing to it
    if (outputStream && outputStream.bytesWritten < MAX_FILE_SIZE) {
        return Promise.resolve(useZip ? zipper : outputStream);
    }
    // Existing stream needs to be finished up if it exists
    if (outputStream) {
        console.log("out file size exceeded " + MAX_FILE_SIZE / 1024 + "KB - rotating...");
        if (useZip) {
            zipper.end();
        }
        else {
            outputStream.end();
        }
    }
    var curOutDir = getOutputDirectoryName();
    console.log("creating new output file in " + curOutDir + "...");
    return new Promise(function (resolve, reject) {
        createOutputDir(curOutDir)
            .then(function () { return getNextFilename(curOutDir); })
            .then(function (filename) {
            outputStream = fs_1.createWriteStream(path_1.join(curOutDir, filename));
            outputStream.on('open', function () {
                console.log("created new output file " + filename + "...");
                if (useZip) {
                    zipper = zlib_1.createGzip();
                    zipper.pipe(outputStream);
                    resolve(zipper);
                }
                else {
                    resolve(outputStream);
                }
            });
            outputStream.on('error', function (err) {
                reject(err);
            });
        });
    });
}
/**
 * Handle process SIGINT event to gracefully end the program and ensure
 * remaing data is written out
 * @return {void}
 */
function onSigint() {
    console.log('shutting process down gracefully. please wait');
    pollers.forEach(function (p) {
        p.stopPolling();
    });
    pqueue.add(function () {
        return getWriteStream();
    })
        .then(function (stream) {
        if (stream === process.stdout) {
            // No need to do any graceful shutdown
            process.exit(0);
        }
        // Wait for the final write(s) to complete after "end" is called
        // and before closing the process completely
        stream.on('close', function () {
            console.log('streams have finished writing. exiting');
            process.exit(0);
        });
        stream.end();
    });
}
function monitor(pids) {
    return connection_1.ensureObdIsConnected()
        .then(function () {
        process.on('SIGINT', onSigint);
        return Promise.map(pids, function (p) {
            var poller = new OBD.ECUPoller({
                pid: p.pid,
                interval: p.interval
            });
            pollers.push(poller);
            poller.on('data', function (output) {
                pqueue.add(function () {
                    return getWriteStream();
                })
                    .then(function (stream) {
                    stream.write(JSON.stringify(output) + "\n");
                });
            });
            poller.startPolling();
        });
    });
}
